"""Core functionalities."""
import logging
import re
from collections import namedtuple

from .decorators import seek_file
from .errors import EncodingError

try:
    from os import scandir
except ImportError:
    from scandir import scandir


logger = logging.getLogger(__name__)

# Official regex from https://www.python.org/dev/peps/pep-0263/
ENCODING_PATTERN = re.compile("^[ \t\v]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)")
ENCODING_LINE = '# -*- coding: {encoding_name} -*-\n'
EncodingInfo = namedtuple('EncodingInfo', 'name lineno error')


def find_encoding(filename):
    """Find information about encoding in file."""
    try:
        with open(filename, 'r') as f:
            encoding_info = _find_file_encoding(f)
            return encoding_info
    except PermissionError:
        logger.error('Cannot open file %s', filename)
        return EncodingInfo('permission denied', -1, EncodingError.PERMISSION_ERROR)
    except FileNotFoundError:
        logger.error('File not found %s', filename)
        return EncodingInfo('no encoding', -1, EncodingError.NOT_FOUND)
    except LookupError as exc:
        msg = str(exc).capitalize()
        logger.warning('%s in %s', msg, filename)
        return EncodingInfo('no encoding', -1, EncodingError.NOT_FOUND)
    except ValueError as exc:
        msg = str(exc).capitalize()
        logger.warning('%s in %s', msg, filename)
        return EncodingInfo('invalid encoding', -1, EncodingError.INVALID_ENCODING)


def _find_file_encoding(f, lineno=1):
    if lineno > 2:
        raise LookupError('encoding not found')

    line = f.readline()
    line_match = re.search(ENCODING_PATTERN, line)
    if line_match:
        encoding_name = validate_encoding(line_match.group(1))
        return EncodingInfo(name=encoding_name, lineno=lineno, error=EncodingError.OK)
    return _find_file_encoding(f, lineno + 1)


def validate_encoding(encoding_name):
    """Validate encoding name."""
    try:
        import codecs
        codecs.lookup(encoding_name)
    except LookupError as e:
        raise ValueError(e)
    return encoding_name


def search_files(path):
    """Find recursively python files in path."""
    files = []
    try:
        for entry in scandir(path):
            if _is_py_file(entry):
                files.append(entry.path)
            elif entry.is_dir():
                subdir_files = search_files(entry.path)
                files.extend(subdir_files)
    except PermissionError as exc:
        logger.error('Cannot open %s', exc.filename)
    except FileNotFoundError:
        logger.error('Directory not found %s', path)
    except NotADirectoryError:
        files.append(path)
    return files


def _is_py_file(entry):
    return entry.is_file() and entry.name.endswith('.py')


def append_encoding(filename, encoding_name, force=False):
    """Append encoding to file."""
    try:
        with open(filename, 'r+') as f:
            _append_file_encoding(f, encoding_name, force)
    except PermissionError:
        logger.error('Cannot open a file %s', filename)
    except FileNotFoundError:
        logger.error('File not found %s', filename)
    except IsADirectoryError:
        logger.error('Not a file %s', filename)


def _append_file_encoding(f, encoding_name, replace=False):
    encoding_name = validate_encoding(encoding_name)
    encoding_line = ENCODING_LINE.format(**locals())
    try:
        encoding_info = _find_file_encoding(f)
    except LookupError:
        lineno = _find_lineno(f)
    else:
        if replace:
            lineno = encoding_info.lineno
        else:
            raise RuntimeError('encoding already exists')

    _write_line(f, lineno, encoding_line, replace)


@seek_file
def _find_lineno(f):
    first_line = f.readline()
    shebang = first_line.startswith('#!')
    return 2 if shebang else 1


@seek_file
def _write_line(f, lineno, content, replace):
    file_content = f.readlines()
    line_index = lineno - 1

    if replace:
        file_content[line_index] = content
    else:
        file_content.insert(line_index, content)

    f.seek(0)
    f.writelines(file_content)



"""Command line interface."""
from contextlib import contextmanager

import click

from . import __version__
from .core import append_encoding, find_encoding
from .core import logger as core_logger
from .core import search_files, validate_encoding
from .errors import error_colour


def validate_encoding_name(ctx, param, value):
    """Validate encoding name option."""
    try:
        return validate_encoding(value)
    except ValueError:
        raise click.BadParameter('unknown encoding')
    except TypeError:
        return None


@click.option('--force', '-f', is_flag=True, help='Overwrite all files encoding')
@click.option('--append', '-A', callback=validate_encoding_name, help='Append encoding to files')
@click.argument('path', type=click.Path(exists=True), default='.')
@click.command()
@click.version_option(version=__version__)
def main(path, append, force):
    """Manage source files encodings."""
    with disable_logger(core_logger):
        for filename in search_files(path):
            if append:
                try:
                    append_encoding(filename, encoding_name=append, force=force)
                except RuntimeError:
                    pass
            encoding = find_encoding(filename)
            click.secho('{0}: {1}'.format(filename, encoding.name), fg=error_colour(encoding.error))


@contextmanager
def disable_logger(logger):
    """Disable logger."""
    logger.disabled = True
    yield logger
    logger.disabled = False



"""Decorators."""
import io
from functools import wraps


def seek_file(func):
    """Seek file passed as first argument."""
    @wraps(func)
    def _wrapper(*args, **kwargs):
        f = args[0]
        if isinstance(f, io.IOBase):
            f.seek(0)
        return func(*args, **kwargs)
    return _wrapper



"""Errors occurred while looking for file encoding."""
import enum


class EncodingError(enum.IntEnum):
    """Encoding error types."""

    OK = 0
    NOT_FOUND = 1
    PERMISSION_ERROR = 2
    INVALID_ENCODING = 3


ERRORS_COLOUR = {
    EncodingError.OK: 'green',
    EncodingError.NOT_FOUND: 'yellow',
    EncodingError.PERMISSION_ERROR: 'red',
    EncodingError.INVALID_ENCODING: 'red',
}


def error_colour(error):
    """Return error's colour."""
    return ERRORS_COLOUR[error]



"""Package configuration."""
import os
import re

from setuptools import find_packages, setup


def find_version(*file_paths):
    """Find version information in file."""
    path = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(path).read()
    version_pattern = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_match = re.search(version_pattern, version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


tests_requirements = [
    'pytest==3.6.0',
    'pylama==7.4.3',
    'pylint==1.9.1',
    'pylama-pylint==3.0.1',
]

setup(
    name='pep263',
    version=find_version('pep263', '__init__.py'),
    author='Bartłomiej Kurzeja',
    url='https://github.com/B3QL/pep263',
    license='MIT',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'click==6.7',
        'colorama==0.3.9',
    ],
    tests_require=tests_requirements,
    extras_require={
        ':python_version<"3.5"': ["scandir==1.7"],
        'test': tests_requirements,
    },
    entry_points={
        'console_scripts': [
            'pep263=pep263.cli:main'
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ],
)
